from sqlalchemy import Column, Integer,String

from xbase_util.db.bean import DbBase


class CurrentConfig(DbBase):
    __tablename__ = 'currentconfig'
    id = Column(Integer, primary_key=True)
    config_id = Column(Integer)
    description = Column(String)


import copy
import re

from scapy.layers.inet import TCP
from scapy.packet import Raw

from xbase_util.xbase_constant import plain_content_type_columns, packetKeyname, src_dst_header, statisticHeader, \
    features_key, plain_body_columns
from xbase_util.xbase_util import firstOrZero


def content_type_is_plain(packet):
    """
    从单个包（包括header和body）中获取content-type并判断是否为可见类型
    :param packet:
    :return:
    """
    if ":" not in packet:
        return False
    for item in packet.replace("-", "_").replace(" ", "").lower().splitlines():
        if "content_type" in item:
            if ":" not in item:
                continue
            content_type = item.split(":")[1].replace("\r", "").strip()
            return content_type in plain_content_type_columns
    return False


def filter_visible_chars(data):
    """
    过滤不可见字符，仅保留可打印的ASCII字符
    :param data:
    :return:
    """
    return ''.join(chr(b) for b in data if 32 <= b <= 126 or b in (9, 10, 13))


def get_all_columns(
        contains_packet_column=False,
        contains_src_dst_column=False,
        contains_statistic_column=False,
        contains_features_column=False,
        contains_plain_body_column=False,
        contains_pcap_flow_text=False
):
    result_columns = []
    if contains_packet_column:
        result_columns += packetKeyname
    if contains_src_dst_column:
        result_columns += src_dst_header
    if contains_statistic_column:
        result_columns += statisticHeader
    if contains_features_column:
        result_columns += features_key
    if contains_plain_body_column:
        result_columns += plain_body_columns
    if contains_pcap_flow_text:
        result_columns.append(contains_pcap_flow_text)
    return result_columns


req_pattern = re.compile(r"(GET|POST|HEAD|PUT|DELETE|OPTIONS|PATCH) \/[^\s]* HTTP\/\d\.\d[\s\S]*?\r\n\r\n",
                         re.DOTALL)
res_pattern = re.compile(r"HTTP/\d\.\d \d{3}.*", re.DOTALL)
req_body_pattern = re.compile(
    r"(GET|POST|HEAD|PUT|DELETE|OPTIONS|PATCH) \/[^\s]* HTTP\/\d\.\d[\s\S]*?(?=HTTP/\d\.\d)", re.DOTALL)

http_version = re.compile(r"HTTP\/(\d\.\d)")
http_req_method = re.compile(r"(GET|POST|HEAD|PUT|DELETE|OPTIONS|PATCH) \/[^\s]* HTTP\/\d\.\d")
http_req_path = re.compile(r"(?:GET|POST|HEAD|PUT|DELETE|OPTIONS|PATCH)\s+(\/[^\s]*)\s+HTTP\/\d\.\d")
res_status_code_pattern = re.compile(r"HTTP\/\d\.\d\s+(\d{3})\s+.*")


# def get_all_packets_by_reg(packets):
#     http_Req_Raw = {}
#     http_methods = ("POST /", "PUT /", "OPTIONS /", "DELETE /", "GET /")
#     for packet in packets:
#         if TCP in packet and Raw in packet:
#             data = packet[Raw].load
#             ack = packet[TCP].ack
#             next_ack = packet[TCP].seq + len(data)
#             if ack not in http_Req_Raw:
#                 http_Req_Raw[ack] = {"time": [float(packet.time)], "data": data, "next_ack": next_ack}
#             else:
#                 http_Req_Raw[ack]["time"].append(float(packet.time))
#                 http_Req_Raw[ack]["data"] += data
#                 http_Req_Raw[ack]["next_ack"] = next_ack
#     packet_list = [
#         {
#             'req_data': item['data'],
#             'res_data': http_Req_Raw[item['next_ack']]['data'],
#             'req_text': filter_visible_chars(item['data']),
#             'res_text': filter_visible_chars(http_Req_Raw[item['next_ack']]['data']),
#             'req_time': item['time'],
#             'res_time': http_Req_Raw[item['next_ack']]['time']
#         }
#         for ack, item in http_Req_Raw.items()
#         if item['next_ack'] in http_Req_Raw  # 检查 next_ack 是否存在
#            and any(method in filter_visible_chars(item['data']) for method in http_methods)
#     ]
#     return packet_list


def get_header_value(header_set, value):
    result = [item for item in header_set if value in item]
    if len(result) != 0:
        return result[0].replace(f"{value}:", "").strip()
    else:
        return ""


def get_detail_by_package(publicField, req_header, req_body, res_header, res_body):
    """
    通过pcap的数量分离session并完善相关字段
    :param publicField: 原始的session单条数据
    :param req_header:请求头
    :param req_body:请求体
    :param res_header:响应头
    :param res_body:响应体
    :return: 完整的单条数据
    """
    res_field = copy.deepcopy(publicField)
    res_field["initRTT"] = firstOrZero(res_field.get("initRTT", 0))
    res_field["length"] = firstOrZero(res_field.get("length", 0))

    http_version_res = http_version.findall(req_header)
    res_field['http.clientVersion'] = http_version_res[0] if len(http_version_res) > 0 else ""
    http_method = http_req_method.findall(req_header)
    http_path = http_req_path.findall(req_header)
    res_field['http.clientVersion'] = http_version_res[0] if len(http_version_res) > 0 else ""
    res_field['http.method'] = http_method[0] if len(http_method) > 0 else ""
    res_field['http.path'] = http_path[0] if len(http_path) > 0 else ""
    request_lines = req_header.splitlines()
    res_field['http.request-referer'] = get_header_value(header_set=request_lines, value="Referer")
    res_field['http.request-content-type'] = get_header_value(header_set=request_lines,
                                                              value="Content-Type")
    res_field['http.hostTokens'] = get_header_value(header_set=request_lines, value="Host")

    res_field['plain_body_src'] = ""
    res_field['plain_body_dst'] = ""
    if content_type_is_plain(req_header):
        res_field['plain_body_src'] = filter_visible_chars(f'{req_body}')
    if content_type_is_plain(res_header):
        res_field['plain_body_dst'] = filter_visible_chars(f'{res_body}')

    http_server_version_res = http_version.findall(res_header)
    res_field['http.serverVersion'] = http_server_version_res[0] if len(http_server_version_res) > 0 else ""

    status_code = res_status_code_pattern.findall(res_header)
    res_field['http.statuscode'] = status_code[0] if len(status_code) > 0 else ""
    response_lines = res_header.splitlines()
    res_field['http.response-server'] = get_header_value(header_set=response_lines, value="Server")
    res_field['http.response-content-type'] = get_header_value(header_set=response_lines,
                                                               value="Content-Type")
    for response in list(set(response_lines + request_lines)):
        key_value = response.replace("\r", "").split(":")
        if len(key_value) == 2:
            key = key_value[0].replace(" ", "").replace("-", "_").lower()
            value = key_value[1].replace(" ", "")
            if f"src_{key}" in src_dst_header:
                res_field[f"src_{key}"] = value
            if f"dst_{key}" in src_dst_header:
                res_field[f"dst_{key}"] = value
    return res_field



import copy
import gzip
import re
import zlib
from io import BytesIO

import numpy as np
from scapy.all import *
from scapy.layers.inet import TCP

REQUEST_LINE_RE = re.compile(rb"^(GET|POST|PUT|DELETE|OPTIONS|HEAD|PATCH)\s[^\r\n]+\r\n", re.MULTILINE)
RESPONSE_LINE_RE = re.compile(rb"^HTTP/\d\.\d\s+\d{3}\s?[^\r\n]*", re.IGNORECASE)


def read_packets(packets):
    last_seq_len = -1
    last_ack = -1
    packet_list = []
    tmp_data = b''
    tmp_packets = []
    for index, pkt in enumerate(packets):
        data = pkt[Raw].load if Raw in pkt else b''
        ack = pkt[TCP].ack
        seq = pkt[TCP].seq
        if seq == last_seq_len:
            # print(f"检测到连续包 数据长度:{len(data)} + seq:{seq}={len(data) + seq}  ack:{ack}")
            tmp_data += data
            tmp_packets.append(pkt)
        elif seq == last_ack:
            if tmp_data != b'':
                if REQUEST_LINE_RE.match(tmp_data) or RESPONSE_LINE_RE.match(tmp_data):
                    packet_list.append({'data': copy.deepcopy(tmp_data), 'pkts': copy.deepcopy(tmp_packets)})
                else:
                    # print("没有新的请求或者响应，就把数据加到上一个里面")
                    if len(packet_list) > 0:
                        # 之前找到过有请求，可以添加到之前的数据，否则说明一开始就没找到请求
                        packet_list[-1]['pkts'].extend(copy.deepcopy(tmp_packets))
                        packet_list[-1]['data'] += tmp_data

            tmp_data = data
            tmp_packets = [pkt]
            # print(f"顺序正确 数据长度:{len(data)} + seq:{seq}={len(data) + seq}  ack:{ack}")
        else:
            # print(f"顺序错误 数据长度:{len(data)} + seq:{seq}={len(data) + seq}  ack:{ack}")
            if len(data) > 0:
                # 但是有数据
                tmp_data += data
                tmp_packets.append(pkt)
        last_ack = ack
        last_seq_len = seq + len(data)
    if tmp_data != b'':
        packet_list.append({'data': copy.deepcopy(tmp_data), 'pkts': copy.deepcopy(tmp_packets)})
        tmp_packets.clear()
    return packet_list

pattern_chuncked = re.compile(rb"Transfer-Encoding:\s*chunked", re.IGNORECASE)
pattern_gzip = re.compile(rb"Content-Encoding:\s*gzip", re.IGNORECASE)
def parse_req_or_res(data, pkts):
    if data.find(b"\r\n\r\n") != -1:
        res = data.split(b"\r\n\r\n", 1)
        header = res[0]
        body = res[1]
    else:
        header = data
        body = b''
    chuncked_pattern = pattern_chuncked.search(header)
    gzip_pattern = pattern_gzip.search(header)
    if chuncked_pattern and b'chunked' in chuncked_pattern.group():
        chunk_lines = [item for item in body.split(b"\r\n") if item != b'']
        data = b''
        next_chunk_size = 0
        for chunk in chunk_lines:
            try:
                next_chunk_size = int(chunk, 16)
                if next_chunk_size == 0:
                    break
                # print(f"接下来的分段大小：{next_chunk_size}")
            except:
                if next_chunk_size > 0:
                    data += chunk
                    # print(f"分段数据大小：{len(data)}")
        result_body = data
    else:
        # print("虽然没有指定chunked，但是我猜出来他就是chunked")
        if body.endswith(b"0\r\n"):
            chunk_lines = [item for item in body.split(b"\r\n") if item != b'']
            data = b''
            next_chunk_size = 0
            for chunk in chunk_lines:
                try:
                    next_chunk_size = int(chunk, 16)
                    if next_chunk_size == 0:
                        break
                    # print(f"接下来的分段大小：{next_chunk_size}")
                except:
                    if next_chunk_size > 0:
                        data += chunk
                        # print(f"分段数据大小：{len(data)}")
            result_body = data
        else:
            result_body = body
    if gzip_pattern and b'gzip' in gzip_pattern.group():
        try:

            decompressed = gzip.decompress(result_body)
            result_body_str = "\n".join(
                [line.strip() for line in decompressed.decode("utf-8", errors="replace").splitlines() if
                 line.strip() != ""])
        except Exception as e:
            result_body_str = result_body.decode("utf-8", errors="replace")
    else:
        result_body_str = result_body.decode("utf-8", errors="replace")
    return header.decode("utf-8", errors="replace"), result_body_str, [float(pkt.time) for pkt in pkts]


def get_all_packets_by_segment(packets):
    res = read_packets(packets)
    request_packets = [item for item in res if REQUEST_LINE_RE.match(item['data'])]
    response_packets = [
        {'first_seq': item['pkts'][0][TCP].seq, 'pkts': item['pkts'], 'first_ack': item['pkts'][0][TCP].ack,
         'data': item['data']} for item in
        res if RESPONSE_LINE_RE.match(item['data'])]
    packet_list = []
    for request in request_packets:
        pkt_list = request['pkts']
        last_pkt = pkt_list[-1]
        # seq = last_pkt[TCP].seq
        ack = last_pkt[TCP].ack
        response = [item for item in response_packets if item['first_seq'] == ack]
        # print(f"找到对应的响应：{len(response)}")
        # print(f"请求：{request['data'].decode('utf-8', errors='replace')}")
        # gzip.decompress(response[0]['data'].split(b'\r\n\r\n')[1])
        # decompressed_data = zlib.decompress((response[0]['data'].split(b'\r\n\r\n')[1])[10:], wbits=zlib.MAX_WBITS | 16)
        if len(response) > 0:
            res_header, res_body, res_times = parse_req_or_res(response[0]['data'], response[0]['pkts'])
            req_header, req_body, req_times = parse_req_or_res(request['data'], request['pkts'])
            packet_list.append({
                "req_header": req_header,
                "req_body": req_body,
                "req_time": req_times,
                "req_packets": len(request['pkts']),
                "res_header": res_header,
                "res_body": res_body,
                "res_time": res_times,
                "res_packets": len(response[0]['pkts']),
            })
        else:
            # print("没响应")
            req_header, req_body, req_times = parse_req_or_res(request['data'], request['pkts'])
            packet_list.append({
                "req_header": req_header,
                "req_body": req_body,
                "req_time": req_times,
                "req_packets": len(request['pkts']),
                "res_header": '',
                "res_body": '',
                "res_time": [],
                "res_packets": 0,
            })
    return packet_list


# if __name__ == '__main__':
    # all_packets = get_all_packets_by_segment(rdpcap("../out/3post.pcap"))
    # res=[
    #     get_detail_by_package({}, package['req_header'], package['req_body'], package['res_header'],
    #                           package['req_body']) for package in all_packets]
    # print(res)


import math
import os
import struct
import time
import zlib
from datetime import datetime
from Crypto.Cipher import AES
from zstandard import ZstdDecompressor


def fix_pos(pos, packetPosEncoding):
    if pos is None or len(pos) == 0:
        return
    if packetPosEncoding == "gap0":
        last = 0
        lastgap = 0
        for i, pos_item in enumerate(pos):
            if pos[i] < 0:
                last = 0
            else:
                if pos[i] == 0:
                    pos[i] = last + lastgap
                else:
                    lastgap = pos[i]
                    pos[i] += last
                last = pos[i]


def group_numbers(nums):
    result = []
    for num in nums:
        if num < 0:
            result.append([num])
        elif result:
            result[-1].append(num)
    return result


def decompress_streaming(compressed_data, id, fro):
    try:
        decompressor = ZstdDecompressor()
        with decompressor.stream_reader(compressed_data) as reader:
            decompressed_data = reader.read()
            return decompressed_data
    except Exception as e:
        print(f"解码错误：{e}  {id}")
        return bytearray()


def read_header(param_map, id):
    shortHeader = None
    headBuffer = os.read(param_map['fd'], 64)
    if param_map['encoding'] == 'aes-256-ctr':
        if 'iv' in param_map:
            param_map['iv'][12:16] = struct.pack('>I', 0)
            headBuffer = bytearray(
                AES.new(param_map['encKey'], AES.MODE_CTR, nonce=param_map['iv']).decrypt(bytes(headBuffer)))
        else:
            print("读取头部信息失败，iv向量为空")
    elif param_map['encoding'] == 'xor-2048':
        for i in range(len(headBuffer)):
            headBuffer[i] ^= param_map['encKey'][i % 256]
    if param_map['uncompressedBits']:
        if param_map['compression'] == 'gzip':
            headBuffer = zlib.decompress(bytes(headBuffer), zlib.MAX_WBITS | 16)
        elif param_map['compression'] == 'zstd':
            headBuffer = decompress_streaming(headBuffer, id, "header")
    headBuffer = headBuffer[:24]
    magic = struct.unpack('<I', headBuffer[:4])[0]
    bigEndian = (magic == 0xd4c3b2a1 or magic == 0x4d3cb2a1)
    # nanosecond = (magic == 0xa1b23c4d or magic == 0x4d3cb2a1)
    if not bigEndian and magic not in {0xa1b2c3d4, 0xa1b23c4d, 0xa1b2c3d5}:
        corrupt = True
        # os.close(param_map['fd'])
        raise ValueError("Corrupt PCAP header")
    if magic == 0xa1b2c3d5:
        shortHeader = struct.unpack('<I', headBuffer[8:12])[0]
        headBuffer[0] = 0xd4  # Reset header to normal
    linkType = struct.unpack('>I' if bigEndian else '<I', headBuffer[20:24])[0]
    return headBuffer, shortHeader, bigEndian


def create_decipher(pos, param_map):
    param_map['iv'][12:16] = struct.pack('>I', pos)
    return AES.new(param_map['encKey'], AES.MODE_CTR, nonce=param_map['iv'])


def read_packet_internal(pos_arg, hp_len_arg, param_map, id):
    pos = pos_arg
    hp_len = hp_len_arg
    if hp_len == -1:
        if param_map['compression'] == "zstd":
            hp_len = param_map['uncompressedBitsSize']
        else:
            hp_len = 2048
    inside_offset = 0
    if param_map['uncompressedBits']:
        inside_offset = pos & param_map['uncompressedBitsSize'] - 1
        pos = math.floor(pos / param_map['uncompressedBitsSize'])
    pos_offset = 0
    if param_map['encoding'] == 'aes-256-ctr':
        pos_offset = pos % 16
        pos = pos - pos_offset
    elif param_map['encoding'] == 'xor-2048':
        pos_offset = pos % 256
        pos = pos - pos_offset

    hp_len = 256 * math.ceil((hp_len + inside_offset + pos_offset) / 256)
    buffer = bytearray(hp_len)
    os.lseek(param_map['fd'], pos, os.SEEK_SET)
    read_buffer = os.read(param_map['fd'], len(buffer))
    if len(read_buffer) - pos_offset < 16:
        return None
    if param_map['encoding'] == 'aes-256-ctr':
        decipher = create_decipher(pos // 16, param_map)
        read_buffer = bytearray(decipher.decrypt(read_buffer))[pos_offset:]
    elif param_map['encoding'] == 'xor-2048':
        read_buffer = bytearray(b ^ param_map['encKey'][i % 256] for i, b in enumerate(read_buffer))[pos_offset:]
    if param_map['uncompressedBits']:
        try:
            if param_map['compression'] == 'gzip':
                read_buffer = zlib.decompress(read_buffer, zlib.MAX_WBITS | 16)
            elif param_map['compression'] == 'zstd':
                read_buffer = decompress_streaming(read_buffer, id, "packet")
        except Exception as e:
            print(f"PCAP uncompress issue:  {pos} {len(buffer)} {read_buffer} {e}")
            return None
    if inside_offset:
        read_buffer = read_buffer[inside_offset:]
    header_len = 16 if param_map['shortHeader'] is None else 6
    if len(read_buffer) < header_len:
        if hp_len_arg == -1 and param_map['compression'] == 'zstd':
            return read_packet_internal(pos_arg, param_map['uncompressedBitsSize'] * 2, param_map, id)
        print(f"Not enough data {len(read_buffer)} for header {header_len}")
        return None
    packet_len = struct.unpack('>I' if param_map['bigEndian'] else '<I', read_buffer[8:12])[
        0] if param_map['shortHeader'] is None else \
        struct.unpack('>H' if param_map['bigEndian'] else '<H', read_buffer[:2])[0]
    if packet_len < 0 or packet_len > 0xffff:
        return None
    if header_len + packet_len <= len(read_buffer):
        if param_map['shortHeader'] is not None:
            t = struct.unpack('<I', read_buffer[2:6])[0]
            sec = (t >> 20) + param_map['shortHeader']
            usec = t & 0xfffff
            new_buffer = bytearray(16 + packet_len)
            struct.pack_into('<I', new_buffer, 0, sec)
            struct.pack_into('<I', new_buffer, 4, usec)
            struct.pack_into('<I', new_buffer, 8, packet_len)
            struct.pack_into('<I', new_buffer, 12, packet_len)
            new_buffer[16:] = read_buffer[6:packet_len + 6]
            return new_buffer
        return read_buffer[:header_len + packet_len]

    if hp_len_arg != -1:
        return None

    return read_packet_internal(pos_arg, 16 + packet_len, param_map, id)


def read_packet(pos, param_map, id):
    if 'fd' not in param_map or not param_map['fd']:
        time.sleep(0.01)
        return read_packet(pos, param_map['fd'], id)
    return read_packet_internal(pos, -1, param_map, id)


def get_file_and_read_pos(id, file, pos_list):
    filename = file['name']
    if not os.path.isfile(filename):
        print(f"文件不存在:{filename}")
        return None
    encoding = file.get('encoding', 'normal')
    encKey = None
    iv = None
    compression = None
    if 'dek' in file:
        dek = bytes.fromhex(file['dek'])
        encKey = AES.new(file['kek'].encode(), AES.MODE_CBC).decrypt(dek)

    if 'uncompressedBits' in file:
        uncompressedBits = file['uncompressedBits']
        uncompressedBitsSize = 2 ** uncompressedBits
        compression = 'gzip'
    else:
        uncompressedBits = None
        uncompressedBitsSize = 0
    if 'compression' in file:
        compression = file['compression']

    if 'iv' in file:
        iv_ = bytes.fromhex(file['iv'])
        iv = bytearray(16)
        iv[:len(iv_)] = iv_
    fd = os.open(filename, os.O_RDONLY)
    param_map = {
        "fd": fd,
        "encoding": encoding,
        "iv": iv,
        "encKey": encKey,
        "uncompressedBits": uncompressedBits,
        "compression": compression,
        "uncompressedBitsSize": uncompressedBitsSize
    }
    res = bytearray()
    headBuffer, shortHeader, bigEndian = read_header(param_map, id)
    res.extend(headBuffer)
    param_map['shortHeader'] = shortHeader
    param_map['bigEndian'] = bigEndian
    # _________________________________
    byte_array = bytearray(0xfffe)
    next_packet = 0
    b_offset = 0
    packets = {}
    i = 0
    for pos in pos_list:
        packet_bytes = read_packet(pos, param_map, id)
        if not packet_bytes:
            continue
        packets[i] = packet_bytes
        while next_packet in packets:
            buffer = packets[next_packet]
            del packets[next_packet]
            next_packet = next_packet + 1
            if b_offset + len(buffer) > len(byte_array):
                res.extend(byte_array[:b_offset])
                b_offset = 0
                byte_array = bytearray(0xfffe)
            byte_array[b_offset:b_offset + len(buffer)] = buffer
            b_offset += len(buffer)
        i = i + 1
    os.close(fd)
    res.extend(byte_array[:b_offset])
    return res


def process_session_id_disk_simple(id, node, packet_pos, esdb, pcap_path_prefix):
    packetPos = packet_pos
    file = esdb.get_file_by_file_id(node=node, num=abs(packetPos[0]),
                                    prefix=None if pcap_path_prefix == "origin" else pcap_path_prefix)
    if file is None:
        return None
    fix_pos(packetPos, file['packetPosEncoding'])
    pos_list = group_numbers(packetPos)[0]
    pos_list.pop(0)
    return get_file_and_read_pos(id, file, pos_list)



from xbase_util.db.bean.FlowBean import FlowBean


class FlowDao:
    def __init__(self,Session):
        self.Session = Session

    def add_flow(self, description, step, flow_id=None):
        with self.Session() as session:
            try:
                if flow_id is None:
                    flow = FlowBean(description=description, step=step)
                    session.add(flow)
                else:
                    flow = session.query(FlowBean).filter_by(id=flow_id).first()
                    flow.description = description
                    flow.step = step
                session.commit()
                return True
            except Exception as e:
                session.rollback()
                print(e)
                return False

    def get_flow_list(self):
        with self.Session() as session:
            try:
                flows = session.query(FlowBean).all()
                return [{
                    'id': item.id,
                    'description': item.description,
                    'step': item.step,
                } for item in flows]
            except Exception as e:
                session.rollback()
                print(e)
                return []

    def delete_by_id(self, id):
        with self.Session() as session:
            try:
                flow = session.query(FlowBean).filter_by(id=id).first()
                if flow:
                    session.delete(flow)
                    session.commit()
                    return True
            except Exception as e:
                session.rollback()
                print(e)
                return False

    def get_flow_by_id(self, id):
        with self.Session() as session:
            try:
                return session.query(FlowBean).filter_by(id=id).first()
            except Exception as e:
                session.rollback()
                print(e)
                return None



import traceback

from xbase_util.db.bean.ConfigBean import ConfigBean


class ConfigDao:
    def __init__(self,Session):
        self.Session = Session

    def add(self, id, description, label_all_true, label_is_output_unmatch,
            label_duration,
            mapping_le_path, pcap_per_subsection, pcap_process, pcap_thread_in_process,
            replace_source, replace_destination, replace_mapping, replace_save_to, session_all_true, session_start_time,
            session_end_time, session_expression, session_alive, session_size, splitNumber,catalogue
            ):
        with self.Session() as session:
            try:
                if id is None:
                    bean = ConfigBean(
                        description=description,
                        label_all_true=label_all_true,
                        label_is_output_unmatch=label_is_output_unmatch,
                        label_duration=label_duration,
                        mapping_le_path=mapping_le_path,
                        pcap_per_subsection=pcap_per_subsection,
                        pcap_process=pcap_process,
                        pcap_thread_in_process=pcap_thread_in_process,
                        replace_source=replace_source,
                        replace_destination=replace_destination,
                        replace_mapping=replace_mapping,
                        replace_save_to=replace_save_to,
                        session_all_true=session_all_true,
                        session_start_time=session_start_time,
                        session_end_time=session_end_time,
                        session_expression=session_expression,
                        session_alive=session_alive,
                        session_size=session_size,
                        splitNumber=splitNumber,
                        catalogue=catalogue
                    )
                    session.add(bean)
                    session.commit()
                    return True
                else:
                    config = session.query(ConfigBean).filter(ConfigBean.id == id).first()
                    if config is None:
                        return False
                    config.description = description
                    config.label_all_true = label_all_true
                    config.label_is_output_unmatch = label_is_output_unmatch
                    config.label_duration = label_duration
                    config.mapping_le_path = mapping_le_path
                    config.pcap_per_subsection = pcap_per_subsection
                    config.pcap_process = pcap_process
                    config.pcap_thread_in_process = pcap_thread_in_process
                    config.replace_source = replace_source
                    config.replace_destination = replace_destination
                    config.replace_mapping = replace_mapping
                    config.replace_save_to = replace_save_to
                    config.session_all_true = session_all_true
                    config.session_start_time = session_start_time
                    config.session_end_time = session_end_time
                    config.session_expression = session_expression
                    config.session_alive = session_alive
                    config.session_size = session_size
                    config.splitNumber = splitNumber
                    config.catalogue=catalogue
                    session.commit()
            except Exception as e:
                session.rollback()
                traceback.print_exc()
                print(e)

    def get_config_file_list(self):
        with self.Session() as session:
            try:
                config_list = session.query(ConfigBean).all()
                return [d.to_dict() for d in config_list]
            except Exception as e:
                session.rollback()
                print(e)

    def remove_by_id(self, id):
        with self.Session() as session:
            try:
                session.query(ConfigBean).filter(ConfigBean.id == id).delete()
                session.commit()
            except Exception as e:
                session.rollback()
                print(e)

    def get_config_by_id(self, id):
        with self.Session() as session:
            try:
                return session.query(ConfigBean).filter(ConfigBean.id == id).first()
            except Exception as e:
                session.rollback()
                print(e)

    def set_config_session_by_id(self, id, session_all_true, session_start_time, session_end_time, session_expression,
                                 session_alive, session_size):
        with self.Session() as session:
            try:
                config = session.query(ConfigBean).filter(ConfigBean.id == id).first()
                config.session_all_true = session_all_true
                config.session_start_time = session_start_time
                config.session_end_time = session_end_time
                config.session_expression = session_expression
                config.session_alive = session_alive
                config.session_size = session_size
                session.commit()
            except Exception as e:
                session.rollback()
                print(e)

    def set_config_pcap_by_id(self, id,
                              pcap_per_subsection,
                              pcap_process,
                              pcap_thread_in_process):
        with self.Session() as session:
            try:
                config = session.query(ConfigBean).filter(ConfigBean.id == id).first()
                config.pcap_per_subsection = pcap_per_subsection
                config.pcap_process = pcap_process
                config.pcap_thread_in_process = pcap_thread_in_process
                session.commit()
            except Exception as e:
                session.rollback()
                print(e)

    def set_config_label_by_id(self, id,
                               label_all_true,
                               label_is_output_unmatch,
                               label_duration):
        with self.Session() as session:
            try:
                config = session.query(ConfigBean).filter(ConfigBean.id == id).first()
                config.label_all_true = label_all_true
                config.label_is_output_unmatch = label_is_output_unmatch
                config.label_duration = label_duration
                session.commit()
            except Exception as e:
                session.rollback()
                print(e)

    def set_config_mapping_by_id(self, id, mapping_le_path):
        with self.Session() as session:
            try:
                config = session.query(ConfigBean).filter(ConfigBean.id == id).first()
                config.mapping_le_path = mapping_le_path
                session.commit()
            except Exception as e:
                session.rollback()
                print(e)

    def set_config_replace_by_id(self, id, replace_source, replace_destination, replace_mapping, replace_save_to):
        with self.Session() as session:
            try:
                config = session.query(ConfigBean).filter(ConfigBean.id == id).first()
                config.replace_source = replace_source
                config.replace_destination = replace_destination
                config.replace_mapping = replace_mapping
                config.replace_save_to = replace_save_to
                session.commit()
            except Exception as e:
                session.rollback()
                print(e)
# 修改脚本带redis获取状态
# 提异常，正常dns和异常都要，要新的黑白样本
# app
# capture
# 日报规范问题 [做了什么][完成的进度][遇到的问题][问题研究的进度和方案]



import json
import logging
import os
import re
from datetime import datetime
from logging.handlers import TimedRotatingFileHandler
from urllib.parse import urlparse, parse_qs

import execjs
import numpy as np
import tldextract
from scapy.layers.dns import DNS

from xbase_util.xbase_constant import parse_path, dns_domain_list


def process_origin_pos(originPos):
    temres = (f"{originPos}"
              .replace("]", "")
              .replace("[", "")
              .replace("\\", "")
              .replace("\"", "")
              .replace(" ", "")
              .split(","))
    return [int(f"{i}") for i in temres]


def parse_expression(expression):
    if expression:
        with open(parse_path, "r") as f:
            ctx = execjs.compile(f.read())
            return eval(ctx.call("parse_exp", expression))
    else:
        return None


def get_cookie_end_with_semicolon_count(text_data):
    count = 0
    for text in text_data.replace("-", "_").lower().splitlines():
        item_text = text.replace("\n", "").replace("\t", "").replace(" ", "")
        if "cookie:" in item_text and f"{item_text}".endswith(";"):
            count = count + 1
    if count == 0:
        return -1
    return count


def get_ua_duplicate_count(text_data):
    ua_list = []
    for text in text_data.replace("-", "_").lower().splitlines():
        item_text = text.replace("\n", "").replace("\t", "").replace(" ", "")
        if "user_agent:" in item_text and f"{item_text}".endswith(";"):
            ua_list.append(item_text.replace("user_agent:", ""))
    count = list(set(ua_list))
    if count == 0:
        return -1
    return sum(count)


def get_res_status_code_list(text_data):
    value_res = []
    res = []
    num_1 = 0
    num_2 = 0
    num_3 = 0
    num_4 = 0
    num_5 = 0

    res.extend([item for item in text_data.splitlines() if item.startswith("HTTP/")])
    for item in res:
        m = re.search(r"\b(\d{3})\b", item)
        if m:
            value_res.append(int(m.group(0)))
    for value in value_res:
        if 0 <= value < 200:
            num_1 = num_1 + 1
        if 200 <= value < 300:
            num_2 = num_2 + 1
        if 300 <= value < 400:
            num_3 = num_3 + 1
        if 400 <= value < 500:
            num_4 = num_4 + 1
        if 500 <= value < 600:
            num_5 = num_5 + 1
    return num_1, num_2, num_3, num_4, num_5


def get_packets_percentage(session, isReq):
    if "source.bytes" in session and "destination.bytes" in session:
        total_bytes = session["source.bytes"] + session["destination.bytes"]
        if total_bytes > 0:
            if isReq:
                return session["source.bytes"] / total_bytes
            else:
                return session["destination.bytes"] / total_bytes
        else:
            return 0.0  # 避免除以0的情况
    else:
        return 0.5


def split_samples(sample, per_subsection):
    num_subsections = len(sample) // per_subsection
    remainder = len(sample) % per_subsection
    subsection_sizes = [per_subsection] * num_subsections
    if remainder > 0:
        subsection_sizes.append(remainder)
        num_subsections += 1
    return num_subsections, subsection_sizes


def split_process(subsection, process_count):
    subsection_per_process = len(subsection) // process_count
    remainder = len(subsection) % process_count
    lengths = []
    start = 0
    for i in range(process_count):
        end = start + subsection_per_process + (1 if i < remainder else 0)
        lengths.append(end - start)
        start = end
    return lengths


def build_es_expression(size, arkime_expression, start_time, end_time, bounded_type="bounded"):
    expression = {"query": {"bool": {"filter": []}}}
    try:
        if size:
            expression['size'] = size

        if bounded_type == "bounded":
            if start_time:
                expression['query']['bool']['filter'].append(
                    {"range": {"firstPacket": {"gte": round(start_time.timestamp() * 1000)}}})
            if end_time:
                expression['query']['bool']['filter'].append(
                    {"range": {"lastPacket": {"lte": round(end_time.timestamp() * 1000)}}})
        elif bounded_type == "last" and start_time and end_time:
            expression['query']['bool']['filter'].append(
                {"range": {"lastPacket": {"gte": round(start_time.timestamp() * 1000),
                                          "lte": round(end_time.timestamp() * 1000)}}})

        arkime_2_es = parse_expression(arkime_expression)
        if arkime_2_es:
            expression['query']['bool']['filter'].append(arkime_2_es)
        return expression
    except Exception as e:
        print(f"请安装nodejs{e}")
        print(arkime_expression)
        exit(1)


def get_uri_depth(url):
    match = re.match(r'^[^?]*', url)
    if match:
        path = match.group(0)
        # 去除协议和域名部分
        path = re.sub(r'^https?://[^/]+', '', path)
        segments = [segment for segment in path.split('/') if segment]
        return len(segments)
    return 0


def firstOrZero(param):
    if type(param).__name__ == 'list':
        if (len(param)) != 0:
            return param[0]
        else:
            return 0
    else:
        return 0


def get_statistic_fields(packets):
    length_ranges = {
        "0_19": (0, 19),
        "20_39": (20, 39),
        "40_79": (40, 79),
        "80_159": (80, 159),
        "160_319": (160, 319),
        "320_639": (320, 639),
        "640_1279": (640, 1279),
        "1280_2559": (1280, 2559),
        "2560_5119": (2560, 5119),
        "more_than_5120": (5120, float('inf'))
    }

    def get_length_range(le):
        for key, (min_len, max_len) in length_ranges.items():
            if min_len <= le <= max_len:
                return key
        return "more_than_5120"

    packet_lengths = {key: [] for key in length_ranges}
    total_length = 0
    packet_len_total_count = len(packets)
    for packet_item in packets:
        length = len(packet_item)
        length_range = get_length_range(length)
        packet_lengths[length_range].append(length)
        total_length += length
    total_time = packets[-1].time - packets[0].time if packet_len_total_count > 1 else 1
    packet_len_average = round(total_length / packet_len_total_count, 5) if packet_len_total_count > 0 else 0
    packet_len_min = min(len(packet_item) for packet_item in packets) if packets else 0
    packet_len_max = max(len(packet_item) for packet_item in packets) if packets else 0
    packet_len_rate = round((packet_len_total_count / total_time) / 1000, 5) if total_time > 0 else 0
    packet_size = [len(p) for p in packets]
    field_map = {
        "packet_size_mean": float(round(np.mean(packet_size), 5)),
        "packet_size_variance": float(round(np.var(packet_size), 5)),
        'packet_len_total_count': packet_len_total_count,
        'packet_len_total_average': packet_len_average,
        'packet_len_total_min': packet_len_min,
        'packet_len_total_max': packet_len_max,
        'packet_len_total_rate': float(packet_len_rate),
        'packet_len_total_percent': 1,
    }
    for length_range, lengths in packet_lengths.items():
        count = len(lengths)
        if count > 0:
            average = round(sum(lengths) / count, 5)
            min_val = min(lengths)
            max_val = max(lengths)
        else:
            average = min_val = max_val = 0
        packet_len_rate = round((count / total_time) / 1000, 5) if total_time > 0 else 0
        percent = round(count / packet_len_total_count, 5) if packet_len_total_count > 0 else 0
        field_map.update({
            f"packet_len_{length_range}_count": count,
            f"packet_len_{length_range}_average": average,
            f"packet_len_{length_range}_min": min_val,
            f"packet_len_{length_range}_max": max_val,
            f"packet_len_{length_range}_rate": float(packet_len_rate),
            f"packet_len_{length_range}_percent": percent
        })
    return field_map


def get_dns_domain(packets):
    domain_name = ""
    for packet_item in packets:
        if DNS in packet_item:
            dns_layer = packet_item[DNS]
            if dns_layer.qd:
                try:
                    domain_name = dns_layer.qd.qname.decode('utf-8')
                    # print(f"dns域名:{domain_name}")
                except Exception:
                    domain_name = str(dns_layer.qd.qname)
                    print(f"dns域名编码失败的字符串:{domain_name}")
                break
    if domain_name.endswith("."):
        domain_name = domain_name[:-1]
    return domain_name


def extract_session_fields(origin_list, geoUtil):
    res = []
    for item in origin_list:
        _source = item.get("_source", {})
        source = _source.get("source", {})
        tcpflags = _source.get("tcpflags", {})
        destination = _source.get("destination", {})
        http = _source.get("http", {})
        dns = _source.get("dns", {})
        tls = _source.get("tls", {})
        uri = http.get('uri', [])
        uri_length = [len(u) for u in uri]
        uri_depth = [get_uri_depth(u) for u in uri]
        uri_filename_length = [get_uri_filename_length(u) for u in uri]
        uri_params = [get_url_param_count(u) for u in uri]
        res.append(geoUtil.get_geo_by_ip({
            "id": item["_id"],
            "node": _source.get("node", ""),
            "segmentCnt": _source.get("segmentCnt", 0),
            "tcpflags.rst": tcpflags.get("rst", 0),
            "tcpflags.ack": tcpflags.get("ack", 0),
            "tcpflags.syn": tcpflags.get("syn", 0),
            "tcpflags.urg": tcpflags.get("urg", 0),
            "tcpflags.psh": tcpflags.get("psh", 0),
            "tcpflags.syn-ack": tcpflags.get("syn-ack", 0),
            "tcpflags.fin": tcpflags.get("fin", 0),
            "source.ip": source.get("ip", ""),
            "destination.ip": destination.get("ip", ""),
            "source.port": source.get("port", ""),
            "source.packets": source.get("packets", ""),
            "source.bytes": source.get("bytes", 0),
            "destination.port": destination.get("port", ""),
            "destination.bytes": destination.get("bytes", 0),
            "destination.packets": destination.get("packets", 0),
            "initRTT": _source.get("initRTT", ""),
            "firstPacket": _source.get("firstPacket", 0),
            "lastPacket": _source.get("lastPacket", 0),
            "ipProtocol": _source.get("ipProtocol", 0),
            "protocolCnt": _source.get("protocolCnt", 0),
            "protocol": _source.get("protocol", []),
            "server.bytes": _source.get("server", {}).get("bytes", 0),
            "totDataBytes": _source.get("totDataBytes", 0),
            "network.packets": _source.get("network", {}).get("packets", 0),
            "network.bytes": _source.get("network", {}).get("bytes", 0),
            "length": _source.get("length", 0),
            "client.bytes": _source.get("client", {}).get("bytes", 0),
            "http.uri": uri,
            "http.uri_length_mean": round(np.nan_to_num(np.mean(uri_length)), 5),
            "http.uri_length_var": round(np.nan_to_num(np.var(uri_length)), 5),
            "http.uri_param_count_mean": round(np.nan_to_num(np.mean(uri_params)), 5),
            "http.uri_param_count_var": round(np.nan_to_num(np.var(uri_params)), 5),
            "http.uri_depth_mean": round(np.nan_to_num(np.mean(uri_depth)), 5),
            "http.uri_depth_var": round(np.nan_to_num(np.var(uri_depth)), 5),
            "http.uri_filename_length_mean": round(np.nan_to_num(np.mean(uri_filename_length)), 5),
            "http.uri_filename_length_var": round(np.nan_to_num(np.var(uri_filename_length)), 5),

            "http.response-content-type": http.get("response-content-type", []),
            "http.bodyMagicCnt": http.get("bodyMagicCnt", 0),
            "http.statuscodeCnt": http.get("statusCodeCnt", 0),
            "http.clientVersionCnt": http.get("clientVersionCnt", 0),
            "http.response-content-typeCnt": http.get("response-content-typeCnt", 0),
            "http.xffIpCnt": http.get("xffIpCnt", 0),
            "http.requestHeaderCnt": http.get("requestHeaderCnt", 0),
            "http.serverVersion": http.get("serverVersion", []),
            "http.serverVersionCnt": http.get("serverVersionCnt", 0),
            "http.responseHeaderCnt": http.get("responseHeaderCnt", 0),
            "http.xffIp": http.get("xffIp", []),
            "http.clientVersion": http.get("clientVersion", []),
            "http.uriTokens": http.get("uriTokens", ""),
            "http.useragentCnt": http.get("useragentCnt", 0),
            "http.statuscode": http.get("statusCode", []),
            "http.bodyMagic": http.get("bodyMagic", []),
            "http.request-content-type": http.get("request-content-type", []),
            "http.uriCnt": http.get("uriCnt", 0),

            "http.useragent": http.get("useragent", ""),
            "http.keyCnt": http.get("keyCnt", 0),
            "http.request-referer": http.get("requestReferer", []),
            "http.request-refererCnt": http.get("requestRefererCnt", 0),
            "http.path": http.get("path", []),
            "http.hostCnt": http.get("hostCnt", 0),
            "http.host": http.get("host", []),
            "http.response-server": http.get("response-server", []),
            "http.pathCnt": http.get("pathCnt", 0),
            "http.useragentTokens": http.get("useragentTokens", ""),
            "http.methodCnt": http.get("methodCnt", 0),
            "http.method": http.get("method", []),
            "http.method-GET": http.get("method-GET", 0),
            "http.method-POST": http.get("method-POST", 0),
            "http.key": http.get("key", []),
            "http.hostTokens": http.get("hostTokens", ""),
            "http.requestHeader": http.get("requestHeader", []),
            "http.responseHeader": http.get("responseHeader", []),

            "dns.ASN": dns.get("ASN", []),
            "dns.RIR": dns.get("RIR", []),
            "dns.GEO": dns.get("GEO", []),
            "dns.alpn": dns.get("https.alpn", []),
            "dns.alpnCnt": dns.get("https.alpnCnt", 0),
            "dns.ip": dns.get("ip", []),
            "dns.ipCnt": dns.get("ipCnt", 0),
            "dns.OpCode": dns.get("opcode", []),
            "dns.OpCodeCnt": dns.get("opcodeCnt", 0),
            "dns.Puny": dns.get("puny", []),
            "dns.PunyCnt": dns.get("puntCnt", 0),
            "dns.QueryClass": dns.get("qc", []),
            "dns.QueryClassCnt": dns.get("qcCnt", 0),
            "dns.QueryType": dns.get("qt", []),
            "dns.QueryTypeCnt": dns.get("qtCnt", 0),
            "dns.status": dns.get("status", []),
            "dns.hostCnt": json.dumps(dns.get("hostCnt", 0)),
            "dns.host": json.dumps(dns.get("host", [])),
            "dns.statusCnt": dns.get("statusCnt", 0),

            "tls.cipher": tls.get("cipher", []),
            "tls.cipherCnt": tls.get("cipherCnt", 0),
            "tls.dstSessionId": tls.get("dstSessionId", []),
            "tls.ja3": tls.get("ja3", []),
            "tls.ja3Cnt": tls.get("ja3Cnt", 0),
            "tls.ja3s": tls.get("ja3s", []),
            "tls.ja3sCnt": tls.get("ja3sCnt", 0),
            "tls.ja4": tls.get("ja4", []),
            "tls.ja4Cnt": tls.get("ja4Cnt", 0),
            "tls.srcSessionId": tls.get("srcSessionId", []),
            "tls.version": tls.get("version", []),
            "tls.versionCnt": tls.get("versionCnt", 0),
            "tls.ja4_r": tls.get("versionCnt", 0),
            "tls.ja4_rCnt": tls.get("versionCnt", 0),
            "packetPos": json.dumps(_source.get("packetPos", [])),
            "traffic_type": item.get("traffic_type", ""),
            "PROTOCOL": item.get("PROTOCOL", ""),
            "DENY_METHOD": item.get("DENY_METHOD", ""),
            "THREAT_SUMMARY": item.get("THREAT_SUMMARY", ""),
            "SEVERITY": item.get("SEVERITY", ""),
        }))
    return res


def get_url_param_count(url):
    query = urlparse(url).query  # 解析 URL 中的查询字符串
    params = parse_qs(query)  # 解析查询字符串为字典
    return len(params)


def get_uri_filename_length(uri):
    match = re.search(r'\.([^./?#]+)$', uri)
    if match:
        extension = match.group(0)
        return len(extension)
    return 0


def get_dns_domain_suffix(domain):
    try:
        for tmp_suffix in dns_domain_list:
            if tmp_suffix in domain:
                return tmp_suffix
        extracted = tldextract.extract(domain)
        return extracted.suffix
    except Exception as e:
        return ""


def check_path(file_path: str):
    directory = os.path.dirname(file_path)
    if not os.path.exists(directory):
        os.makedirs(directory)
    return file_path


def get_project_root(project_name):
    current_directory = os.getcwd()
    while current_directory != os.path.dirname(current_directory):
        if os.path.basename(current_directory) == project_name:
            return f"{current_directory}{os.sep}"
        current_directory = os.path.dirname(current_directory)
    return None


def setup_logger(process_name, dir):
    logger = logging.getLogger(process_name)
    logger.setLevel(logging.DEBUG)  # 设置日志级别
    log_filename = check_path(f"{dir}/{process_name}.log")
    handler = TimedRotatingFileHandler(
        log_filename,
        when="midnight",
        interval=1,
        backupCount=30,
        encoding="utf-8"
    )

    def custom_namer(filename):
        base, ext = os.path.splitext(os.path.splitext(filename)[0])
        current_time = datetime.now().strftime("_%Y_%m_%d")
        return f"{base}{current_time}{ext}"

    handler.namer = custom_namer
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger
